/*
     CROSSO - crossword puzzle generator (c) J. Schulenburg

   v0.2: segfault fixed
      
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


const char *version_string="CROSSO v0.2 (c) jschulen@gmx.de, Nov 2002";

int help(){  /* this is also a documentation for developper */
  printf(
   "%s\n"
   "--- the crossword puzzle generator ---\n"
   "options:\n"
   " -q   be quiet\n"
   " -r#  initialize random generator              (default is 0)\n"
   " -o#  set output 0=text, 1=xfig, 2=xml, 3=dbg  (default is 0)\n"
   " -l#  set optimization loops, higher is better (default is 1, range 1..32768)\n"
   " -x#  set number of columns (default is -x22, range 4..181)\n"
   " -y#  set number of columns (default is -y17, range 4..181)\n"
   " -f*  set database filename (default is -flist1.dat)\n",version_string);
   /* range maximums are not hard, but fit to a 32bit-integer range */
  return(0);
}

/* --------------- declarations --------------------------- */

#define MAXNUM 1024 /* max number of entries per [re]malloc */

int verbose=1;  /* debug mode */

enum direction { horizontal, vertical };

struct sentry {
  int dx, dy; /* position for description */
  int sx, sy; /* start of word */
  int direct;
  char descr[128]; /* description (ToDo: replace fixed size) */
  char word[128];  /* word/entry */
};

struct spuzzle {
   int x, y;  /* dimension */
   char *aa; /* array[x*y], ' '=free, '*'=free for descr., '#'=descr, A-Z:word
      ToDo: change to wchar_t for international version */
   int num;             /* number of entries */
   struct sentry *elist; /* list of entries */
};

struct spuzzle pa={ 21, 17, NULL, 0, NULL },
               pb={ 21, 17, NULL, 0, NULL };

/* --------------- output functions --------------------------- */

/* split description in 1..4 lines of max 10 width */ 
int split_descr( char *descr, char split[4][11]){
  int i,j,ij;
  for(i=0;i<4;i++) for(j=0;j<11;j++) split[i][j]=0;
  for(i=j=ij=0;i<4 && descr[ij];ij++) {
    if (j>9) { i++; j=0; if (i>3) break; }
    if (descr[ij]=='\\' && descr[ij+1]=='-') {
      split[i][j]='-'; ij++; j=0; i++; continue; } 
    if (descr[ij]=='\\' && descr[ij+1]==' ') {
      ij++; j=0; i++; continue; }
    split[i][j]=descr[ij]; j++;
  }
  return 0;
};

/* output to my favorite format */
int out_xfig( struct spuzzle *pp ){
  int i,j; char cc,split[4][11];
  struct sentry *p;
  /* output matrix */
  printf("#FIG 3.2\nLandscape\nCenter\nMetric\nA4\n100.00\nSingle\n"
         "-2\n1200 2\n# created by %s\n",version_string);
  for ( i=0; i<pp->y; i++) {
    for ( j=0; j<pp->x; j++) {
      cc=pp->aa[pp->x*i+j];
      /* draw boxes, units are mm*100.00 */ 
      if (cc=='*' || cc=='#' || cc==' ')
      printf("# 10%/5 filled square\n"
             "2 2 0 1 0 0 50 0 2 0.000 0 0 -1 0 0 5\n"
             "\t %5d %5d  %5d %5d  %5d %5d  %5d %5d  %5d %5d\n",
              j   *1000, i   *1000,(j+1)*1000, i   *1000,
             (j+1)*1000,(i+1)*1000, j   *1000,(i+1)*1000,
              j   *1000, i   *1000);
      else
      printf("# empty square\n"
             "2 2 0 1 0 7 50 0 -1 0.000 0 0 -1 0 0 5\n"
             "\t %5d %5d  %5d %5d  %5d %5d  %5d %5d  %5d %5d\n",
              j   *1000, i   *1000,(j+1)*1000, i   *1000,
             (j+1)*1000,(i+1)*1000, j   *1000,(i+1)*1000,
              j   *1000, i   *1000);
    } 
  }
  /* output list of questions */
  for ( j=0; j<pp->num; j++) {
      p=&(pp->elist[j]);
      printf("# text at %d %d for %d %d direction %d\n",
                p->dx,p->dy,p->sx,p->sy,p->direct);
      /* split description to fit in a square */
      split_descr(p->descr,split);
      /* #6=font:22=Helvetica_narrow_bold, #7=fontsize
       * #10=y-size(ca.(fontsize+1)*10), #11=x-size
       */
      /* ToDo: fit the description to the box */
      for (i=0; i<4 && split[i][0]; i++)
        printf("4 0 0 40 0 22 14 0.0000 4 150 620 %5d %5d %s\\001\n",
                p->dx*1000+50,p->dy*1000+220+i*220,split[i]);

      /* draw pointer */
      if (p->direct==0 && p->sx>p->dx)
        printf("# point right\n"
               "2 3 0 1 0 0 50 0 20 0.000 0 0 -1 0 0 4\n"
               "\t %5d %5d  %5d %5d  %5d %5d  %5d %5d\n",
               p->sx*1000     ,p->sy*1000+ 400,
               p->sx*1000+ 100,p->sy*1000+ 500,
               p->sx*1000     ,p->sy*1000+ 600,
               p->sx*1000     ,p->sy*1000+ 400);
      if (p->direct==0 && p->sy!=p->dy)
        printf("# point up/down right\n"
               "2 1 0 1 0 7 50 0 -1 0.000 0 0 -1 0 0 2\n"
               "\t %5d %5d  %5d %5d\n"
               "2 3 0 1 0 0 50 0 20 0.000 0 0 -1 0 0 4\n"
               "\t %5d %5d  %5d %5d  %5d %5d  %5d %5d\n",
               p->sx*1000+ 100,p->sy*1000+ 500+ 500*(p->dy-p->sy),
               p->sx*1000+ 100,p->sy*1000+ 500,
               p->sx*1000+ 100,p->sy*1000+ 400,
               p->sx*1000+ 200,p->sy*1000+ 500,
               p->sx*1000+ 100,p->sy*1000+ 600,
               p->sx*1000+ 100,p->sy*1000+ 400);
      if (p->direct==1 && p->sy>p->dy)
        printf("# point down\n"
               "2 3 0 1 0 0 50 0 20 0.000 0 0 -1 0 0 4\n"
               "\t %5d %5d  %5d %5d  %5d %5d  %5d %5d\n",
               p->sx*1000+ 400,p->sy*1000     ,
               p->sx*1000+ 500,p->sy*1000+ 100,
               p->sx*1000+ 600,p->sy*1000     ,
               p->sx*1000+ 400,p->sy*1000     );
      if (p->direct==1 && p->sx!=p->dx)
        printf("# point right/left down\n"
               "2 1 0 1 0 7 50 0 -1 0.000 0 0 -1 0 0 2\n"
               "\t %5d %5d  %5d %5d\n"
               "2 3 0 1 0 0 50 0 20 0.000 0 0 -1 0 0 4\n"
               "\t %5d %5d  %5d %5d  %5d %5d  %5d %5d\n",
               p->sx*1000+ 500+ 500*(p->dx-p->sx),p->sy*1000+ 100,
               p->sx*1000+ 500,p->sy*1000+ 100,
               p->sx*1000+ 400,p->sy*1000+ 100,
               p->sx*1000+ 500,p->sy*1000+ 200,
               p->sx*1000+ 600,p->sy*1000+ 100,
               p->sx*1000+ 400,p->sy*1000+ 100);
  }
  return 0;
}

/* print out text version of the puzzle and list of descriptions */
int out_text( struct spuzzle *pp ){
  int i,j; int *aa; char cc;
  struct sentry *p;
  aa=(int *)malloc(pp->x * pp->y * sizeof(int)); if(!p) exit(1);
  /* initialize matrix */
  for ( i=0; i<pp->y; i++) {
    for ( j=0; j<pp->x; j++) {
      aa[pp->x*i+j]=0;
    }
  }
  /* fill matrix from list */
  for ( j=0; j<pp->num; j++) {
      p=&(pp->elist[j]);
      aa[pp->x * p->sy + p->sx]=j+1;
  }
  /* output matrix */
  for ( i=0; i<pp->y; i++) {
    for ( j=0; j<pp->x; j++) printf("+----"); printf("+\n");
    for ( j=0; j<pp->x; j++) {
      cc=pp->aa[pp->x*i+j]; if(cc=='*' || cc==' ') cc='#';
      if (aa[pp->x*i+j]) printf("|%3d ",aa[pp->x*i+j]);
      else if (cc=='#')  printf("|####");
      else               printf("|    ");
    } printf("|\n");
    for ( j=0; j<pp->x; j++) {
      cc=pp->aa[pp->x*i+j]; if(cc=='*' || cc==' ') cc='#';
      if (cc=='#') printf("|####");
      else         printf("|    ");
    } printf("|\n");
  }
  for ( j=0; j<pp->x; j++) printf("+----"); printf("+\n");
  /* output list of questions */
  for ( j=0; j<pp->num; j++) {
      p=&(pp->elist[j]);
      printf("%3d %c : %s\n",j+1,((p->direct)?'V':'H'),p->descr);
  }
  free(aa);
  return 0;
}

/* print out text version of solved puzzle and list of descriptions */
int dbg_out( struct spuzzle *pp ){
  int i,j; struct sentry *p;
  fprintf(stderr,"# Output puzzle, dim=%d*%d\n",pp->x,pp->y);
  if (pp->aa) {
    for ( i=0; i<pp->y; i++) {
      fprintf(stderr,".");
      for ( j=0; j<pp->x; j++) {
        fprintf(stderr," %c ",pp->aa[pp->x*i+j]);
      } fprintf(stderr,".\n");
    }
  }
  fprintf(stderr,"# entries: %d\n",pp->num);
  if (pp->elist) {
    for ( j=0; j<pp->num; j++) {
      p=&(pp->elist[j]);
      fprintf(stderr,"%3d: %2d,%2d  %2d,%2d  %d %s : %s\n",j+1,
        p->sx,p->sy,p->dx,p->dy,p->direct,p->word,p->descr);
    }
  }
  return 0;
}

/* ToDo: Is this XML? Feel free and help me to improve this code. */
int out_xml( struct spuzzle *pp ){
  int i,j; char cc; struct sentry *p;
  printf("<!-- Output from %s (crossword puzzle) -->\n",version_string);
  printf(" <array size=%d,%d>\n",pp->x,pp->y);
  for ( i=0; i<pp->y; i++) {
    for ( j=0; j<pp->x; j++) {
      cc=pp->aa[pp->x*i+j];
      if (cc==' ') printf("  <field pos=%d,%d style=\"unused\" />\n",j,i);
      if (cc=='#') printf("  <field pos=%d,%d style=\"description\" />\n",j,i);
      if (cc=='*') printf("  <field pos=%d,%d style=\"block\" />\n",j,i);
    }
  }
  printf(" <array/>\n");
  printf(" <entries num=%d>\n",pp->num);
  for ( j=0; j<pp->num; j++) {
    p=&(pp->elist[j]);
    printf("  <entry idx=%d start=%d,%d dpos=%d,%d hv=%d word=\"%s\" descr=\"%s\" />\n",j+1,
      p->sx,p->sy,p->dx,p->dy,p->direct,p->word,p->descr);
  }
  printf(" </entries>\n");
  return 0;
}

/* --------------- generator functions --------------------------- */

int reset_puzzle( struct spuzzle *pp ){
  int i,j;
  for ( i=0; i<pp->y; i++ )
  for ( j=0; j<pp->x; j++ ) pp->aa[pp->x*i+j]=' ';
  pp->num=0;
  return 0;
}

/* copy from pa to pb */
int copy_puzzle( struct spuzzle *pa, struct spuzzle *pb ){
  int i,j;
  if (pb->x!=pa->x || pb->y!=pa->y) fprintf(stderr,"error L%d\n",__LINE__);
  memcpy(pb->aa   ,pa->aa   ,pa->x * pa->y * sizeof(char));
  memcpy(pb->elist,pa->elist,MAXNUM * sizeof(struct sentry));
  pb->x=pa->x;
  pb->y=pa->y;
  pb->num=pa->num;
  return 0;
}

int ini_puzzle( struct spuzzle *pp ){
  if (pp->aa) free(pp->aa);  /* avoid memory leak */
  pp->aa = (char *) malloc(pp->x * pp->y * sizeof(char));
  
  /* ToDo: free elements if pp->num>0 and pp->elist */
  if (pp->elist) free(pp->elist);  /* avoid memory leak */
  pp->elist = (struct sentry *) malloc(MAXNUM * sizeof(struct sentry));

  reset_puzzle( pp );
  return 0;
}

int free_puzzle( struct spuzzle *pp ){
  if (pp->aa) free(pp->aa);  /* avoid memory leak */
  if (pp->elist) free(pp->elist);  /* avoid memory leak */
  pp->aa=NULL;
  pp->elist=NULL;
  pp->num=0;
  return 0;
}


/* ToDo: finish this work */
/* check if word fits to the puzzle, return value (#crossses)
   in:  fit.word - word to check
        pp - puzzle
   out: fit.dx,dy,sx,sy,direct
   ret: -1 no fit, >=0 num_crosses
 */
int fit_word( struct sentry *fit, struct spuzzle *pp ) {
  int bx,by,bdx,bdy,x,y,tx,ty,id,sx,sy,dx,dy,i,i1,wlen,crosses,bcrosses;
  int d,bd;
  char cc;
  bdx=bdy=bx=by=bcrosses=-1; /* mark as "not found" */
  for ( wlen=0; fit->word[wlen]; wlen++ ) ;
  /* check if entry is unique (can alternaively made externaly) */
  for (i=0; i<pp->num; i++){
    if (strcasecmp(fit->word,pp->elist[i].word)==0) break;
  } if (i<pp->num) return -1;
  /* check the array */
  for ( id=0;id<2; id++ )
  for ( x=0; x<pp->x; x++)
  for ( y=0; y<pp->y; y++)
  {
    /* first row (or column) seems to be a bad starting point
     * for horizontal (or vertical) words */
    if ((y==0 && id==0) || (x==0 && id==1)) continue;
    d=((id)?vertical:horizontal);
    if (d==horizontal && x+wlen>pp->x) continue; /* to long */
    if (d==vertical   && y+wlen>pp->y) continue;

    sx=x; sy=y; if(d==horizontal) sx--; else sy--;
    if (sx>=0 && sy>=0){ 
      cc=pp->aa[sy*pp->x+sx];
      /* test if start of word is available */
      if (cc!='#' && cc!='*' && cc!=' ') continue; /* used by char? */
    }

    dx=x; dy=y; /* check for free descr-field (3 possibilities) */
    for ( i=0; i<4; i++){
      /* first check the field in front of the word */
      if (i==(0^id<<1)) { dx=x-1; dy=y; if (dx<0) continue; }
      if (i==(1^id<<1)) { dx=x+1; dy=y; if (dx>=pp->x || d==horizontal) continue; }
      if (i==(2^id<<1)) { dx=x; dy=y-1; if (dy<0) continue; }
      if (i==(3^id<<1)) { dx=x; dy=y+1; if (dy>=pp->y || d==vertical) continue; }
      cc=pp->aa[dy*pp->x+dx];
      if (cc!='*' && cc!=' ') continue; /* used by char or descr? */
      break; /* have found a place for description */
    } if (i==4) continue; /* no free descr-field found */

    crosses=0;
    tx=x; ty=y;
    for ( i1=0; fit->word[i1]; i1++ ) {
      cc=pp->aa[pp->x*ty+tx];
      if (cc!=fit->word[i1] && cc!=' ') break;
      if (cc==fit->word[i1]) crosses++;
      if (d==horizontal) tx++;
      else               ty++;
    } 
    if (i1<wlen) continue; /* does not fit */
    /* ToDo: avoid subwords (ex: BEST within ASBEST) */
    if (tx<pp->x && ty<pp->y && pp->aa[pp->x*ty+tx]!=' '
                             && pp->aa[pp->x*ty+tx]!='*'
                             && pp->aa[pp->x*ty+tx]!='#')
      continue;
    if (bcrosses>0) {
      if (crosses<1) continue; /* we need at least one crossing */
    }
    /* ToDo: store as best in fit */
    bx=x; by=y; bdx=dx; bdy=dy; bd=d; bcrosses=crosses;
    if (crosses<1 && pp->num>0) continue;
    /* do not break, if crossing is possible */
    if (random()%(pp->x*pp->y)>0) { id=2; x=pp->x; y=pp->y; break; }
  }
  if (bx<0) return -1; /* do not fit */
  fit->dx=bdx;
  fit->dy=bdy;
  fit->sx=bx;
  fit->sy=by;
  fit->direct=bd;
  return (bcrosses);
}

/* parse data-line from file and put to fit, return length of word */
int get_data( char *buf, int len, struct sentry *fit){
  int nl,w1,i1,x;
  nl=strlen(buf);
  if (nl<1) return 0;  /* no data */
  w1=strcspn(buf," \t\r\n,;.=-/_<>#%");
  if (w1<1 || w1>126) return 0; /* skip comments <!-- --> */
  strncpy(fit->word,buf,w1);
  for (i1=w1;i1<128;i1++) fit->word[i1]=0; /* better dgb */
  for (i1=w1;i1<128;i1++) if (buf[i1]=='"') break;
  for (x=0,i1++;i1<127;i1++,x++) { /* copy "description" */
     if (buf[i1]=='"') break;
     fit->descr[x]=buf[i1]; /* better dgb */
  }
  for (   ;x<128;x++) fit->descr[x]=0; /* fill zeros */
  return w1;
}

/* read data from file, return number of chars */
/* ToDo:
 * - map? the file and generate an index for faster access
 * - mix the words (indexes) randomly
 */
int generate_puzzle( struct spuzzle *pp, char *fname ) {
  int i1,ni,v1,v2,x,y,num_chars;
  FILE *f1; char buf[128];
  struct sentry fit, bfit;
  num_chars=0;
  for(;;){ /* fill the puzzle */
    ni=0; v2=-1;    /* count up to 4(?) possibilities */
    /* select ni=1-4(?) fitting nonused random words, abort if 0 */
    f1=fopen(fname,"rt"); if (!f1) perror("fopen");
    for(;;){  /* number of fitting words */
      for (i1=0;i1<128;i1++) buf[i1]=0; /* 1..127 for better debugging */
      if (!fgets(buf,128,f1)) break; /* read 0..127 chars incl. \n and add \0 */
      if (!get_data(buf,128,&fit)) continue;

      /* check if word fits else try next, v1=crosses */
      if ((v1=fit_word(&fit,pp))<0) continue;

      if (v1<1 && pp->num>0) continue; /* no uncrossed words allowed */
      /* give every fitting word the same chance */
      if (ni>0 && strlen(bfit.word)>strlen(fit.word) && (random()&15)>1) continue; else 
      if (ni>0 && (random()%ni>0)) continue;
      /* replace the choice for next word */
      memcpy(&bfit,&fit,sizeof(struct sentry)); /* fit => bfit */
      ni++; v2=v1; /* count words and crossings */
    }
    fclose(f1);
//    fprintf(stderr," num=%d ni=%d v2=%d\n",pp->num,ni,v2);
    /* put best fitting to array and repeat */
    if (!ni) break;  /* no fitting word */
    /* add the best word to the list and the array and repeat */
    if (pp->num>=MAXNUM) break;
    memcpy(&(pp->elist[pp->num]),&bfit,sizeof(struct sentry));
    /* put to the array, ToDo: move to a subfunction */
    x=bfit.sx; y=bfit.sy;
    pp->aa[bfit.dy*pp->x+bfit.dx]='#'; /* mark as description */
    for ( i1=0; bfit.word[i1]; i1++ ) {
      pp->aa[pp->x*y+x]=bfit.word[i1];
      if (bfit.direct==horizontal) x++;
      else                         y++;
    } num_chars+=i1; /* this is the quality of the puzzle */
    /* mark end of word */
    if (x<pp->x && y<pp->y && pp->aa[pp->x*y+x]==' ')
      pp->aa[pp->x*y+x]='*'; /* free for descr. */
    /* mark start of word */
    x=bfit.sx; y=bfit.sy; if (bfit.direct==horizontal) x--; else y--;
    if (x>0 && y>0 && pp->aa[pp->x*y+x]==' ')
      pp->aa[pp->x*y+x]='*'; /* free for descr. */
    pp->num++;
  }
  return num_chars;
}

/* --------------- main function --------------------------- */

int main(int argn, char *argv[]){
  int i,j,l,ll=1,oo=0;
  char *fname="list1.dat";
  for ( i=1; i<argn; i++ ) {
    if (argv[i][0]=='-') {
      switch ( argv[i][1] ) {
        case 'h': help(); exit(0); break;
        case 'q': verbose=0; break;
        case 'f': fname=argv[i]+2; break;
        case 'r': srand(atoi(argv[i]+2)); break;
        case 'o': oo=atoi(argv[i]+2); break;
        case 'l': ll=atoi(argv[i]+2); break;
        case 'x': j=atoi(argv[i]+2); 
         if (j>2) pb.x=pa.x=j;
         else fprintf(stderr, "ERR: invalid x value\n");
         break;
        case 'y': j=atoi(argv[i]+2); 
         if (j>2) pb.y=pa.y=j;
         else fprintf(stderr, "ERR: invalid y value\n");
         break;
        default:
         fprintf(stderr,"ERR: unknown option %s, try -h for help\n",argv[i]);
         exit(1);
      }
    }
  }
  ini_puzzle(&pa);
  ini_puzzle(&pb);
  for (j=-1,l=0;l<ll;l++) { /* repeat for good results */
    reset_puzzle(&pa);
    i=generate_puzzle(&pa,fname);
    if ((oo&3)==3) dbg_out(&pa);
    if (verbose) fprintf(stderr,"%4d: num_c=%4d %4d\r",l,i,j);
    if (i>j) { j=i; copy_puzzle(&pa,&pb); }
  }
  if (verbose) fprintf(stderr," --- best found result ---\n");
  if ((oo&3)==3) dbg_out(&pb);
  if (verbose) fprintf(stderr," num_c=%4d\n\n",j);
  if ((oo&3)==0) out_text(&pb);
  if ((oo&3)==1) out_xfig(&pb);
  if ((oo&3)==2) out_xml(&pb);
  free_puzzle(&pa);
  free_puzzle(&pb);
  return 0;
}
